(*  title     : A Shallow encoding of ITL in Isabelle/HOL
    Authors   : Antonio Cau     <cau.researcher at gmail.com>
                Ben Moszkowski
                David Smallwood <drs at dmu.ac.uk>
    Maintainer: Antonio Cau     <cau.researcher at gmail.com>        
    License   : BSD
*)
section \<open>Finite ITL Examples\<close>

theory Example
imports
   FOTheorems TimeReversal
begin


(*
 sledgehammer_params [minimize=true,preplay_timeout=10,timeout=60,verbose=true,
                    provers="cvc4 e z3 vampire spass  " ]
*)
(*
declare [[show_types]]
*)
(*
declare [[show_Consts]]
*)

subsection \<open>Example 1\<close>

definition F1 :: "nat statefun \<Rightarrow> temporal "
where " F1 w \<equiv> TEMP \<box> ( #0 \<le> $w )" 

definition Init1 :: "nat statefun \<Rightarrow> temporal "
where "Init1 w \<equiv> TEMP $w = #0"

lemma init1:
 "(\<langle>s0,s1,s2\<rangle> \<Turnstile> len(2) \<and> Init1 w) = ((w s0) = 0) "
by (simp add: Init1_def itl_defs)

lemma  exist_test_F1 :
 "\<turnstile> \<exists>\<exists> w. F1 w"
proof -
 have 1: "\<And> w. \<turnstile> F1 w" by (simp add: itl_defs  F1_def Valid_def)
 from 1 show ?thesis by (meson EExI MP)
qed
 

subsection \<open>Example 2\<close>

locale Test =
 fixes v :: "state \<Rightarrow> nat"
 fixes v1 :: "state \<Rightarrow> nat"
 fixes y :: "state \<Rightarrow> bool" 
 fixes z :: "state \<Rightarrow> int"
 fixes F2 :: "nat statefun \<Rightarrow> temporal "
 fixes F3 :: "bool statefun \<Rightarrow> temporal " 
 fixes F4 :: "int statefun \<Rightarrow> temporal" 
 fixes F5 :: "nat statefun \<Rightarrow> temporal "
 fixes Init2 :: "nat statefun \<Rightarrow> temporal "
 fixes Init3 :: "bool statefun \<Rightarrow> temporal " 
 defines "F2 \<equiv> (\<lambda> v. TEMP \<box> ( #0 \<le> $v )) "
 defines "F3 \<equiv> (\<lambda> p. TEMP \<box> ( $p \<or> \<not> $p))" 
 defines "F4 \<equiv> (\<lambda> z. TEMP \<box> ( #0\<le> $z \<or> $z < #0))"  
 defines "F5 \<equiv> (\<lambda> v. TEMP $v=#0 \<and> v gets $v+#1)"
 defines "Init2 \<equiv> (\<lambda> v. TEMP $v = #0)"
 defines "Init3 \<equiv> (\<lambda> p. TEMP $p)"


lemma (in Test) currentval_test :
 " (s \<Turnstile> ($v = #0)) = ( (v (inth s 0)) = 0)"
by (simp add: itl_def)

lemma (in Test) nextempty_test :
 " (\<langle>s0\<rangle> \<Turnstile> v$) =  (\<some> x. x=x)"
by (simp add: itl_def)

lemma (in Test) nextempty_test_1 :
 " (\<langle>s0\<rangle> \<Turnstile> v$ = v$)"
by simp

lemma (in Test) nextempty_test_2 :
 " (\<langle>s0\<rangle> \<Turnstile> v$ = v1$)"
by (simp add: Test.nextempty_test)

lemma (in Test) nextcurrent_test:
 " (\<langle>s0,s1\<rangle> \<Turnstile> skip \<and> ($v=#0) \<and> (v$=$v+#1)) = (((v s0) = 0) \<and> ((v s1) = 1  ))  "
unfolding itl_defs by auto

lemma (in Test) nextcurrentfinpenult_test:
 " (\<langle>s0,s1,s2,s3\<rangle> \<Turnstile> len(3) \<and> v =: !v-#1 \<and> v \<leftarrow> #3 \<and> $v=#0 \<and> v := $v+#1 ) = 
     (((v s0) = 0) \<and> ((v s1) = 1 \<and> (v (s2)) = 2 \<and> ((v s3) = 3   )))  "
by (simp add: itl_def itl_defs)
 (metis One_nat_def add_Suc_shift add_diff_cancel_right' le_add1 min.orderE min_0R numeral_2_eq_2 
    numeral_3_eq_3 plus_1_eq_Suc zero_neq_one) 



lemma (in Test) stable_test:
 "(\<langle>s0,s1,s2,s3\<rangle> \<Turnstile> len(3) \<and> stable v \<and> $v=#0) = 
   ((v s0) = 0 \<and> (v s1) = 0 \<and> (v s2) = 0 \<and> (v s3) = 0)"
using nat_le_linear by (simp add: itl_defs Nitpick.case_nat_unfold, auto)


lemma (in Test) revnextcurrentfinpenult_test:
 " (\<langle>s0,s1,s2,s3\<rangle> \<Turnstile> (len 3 \<and> v! = !v-#1 \<and> !v = #3 \<and> $v=#0 \<and> v$=$v+#1)\<^sup>r) = 
     (((v s3) = 0) \<and> ((v s2) = 1 \<and> (v (s1)) = 2 \<and> ((v s0) = 3   )))  "
unfolding reverse_d_def len_defs current_val_d_def next_val_d_def 
          penult_val_d_def fin_val_d_def by auto

lemma (in Test) exist_test_F2 :
 "\<turnstile> \<exists>\<exists> v. F2 v"
proof -
 have 1: "\<turnstile> F2 v" by (simp add: itl_defs F2_def Valid_def)
 from 1 show ?thesis by (meson EExI MP)
qed

lemma (in Test) exist_test_F3 :
 "\<turnstile> \<exists>\<exists> y. F3 y"
proof -
 have 1: "\<turnstile> F3 y" by (simp add: itl_defs  F3_def Valid_def)
 from 1 show ?thesis by (meson EExI MP)
qed





subsection \<open>Example 3\<close>

locale Test1 =
 fixes v :: "state \<Rightarrow> nat"
 fixes F5 :: "nat statefun \<Rightarrow> nat \<Rightarrow> temporal "
 defines "F5 \<equiv> (\<lambda> v n. TEMP $v=#0 \<and> v gets $v+#1 \<and> fin($v=#n))"


lemma (in Test1) test_E_F5_1:
 "(
           x ( inth w (0::nat)) = (0::nat) \<and>
           (\<forall>i<ilen w. x ( inth w (Suc i)) = Suc (x ( inth w i))) \<and>
           x ( inth w (ilen w)) = n) \<longrightarrow>
   (
           x ( inth w (0::nat)) = (0::nat) \<and>
           (\<forall>i\<le>ilen w. x ( inth w (i)) = i) \<and>
           x ( inth w (ilen w)) = n)"
proof auto
 show " \<And>i. x ( inth w 0) = 0 \<Longrightarrow>
         \<forall>i<ilen w. x ( inth w (Suc i)) = Suc (x ( inth w i)) \<Longrightarrow>
         n = x ( inth w (ilen w)) \<Longrightarrow> i \<le> ilen w \<Longrightarrow> x ( inth w i) = i"
  proof -
   fix i 
   assume 0: "x ( inth w 0) = 0"
   assume 1: "\<forall>i<ilen w. x ( inth w (Suc i)) = Suc (x ( inth w i))"
   assume 2: "n = x ( inth w (ilen w))"
   assume 3: "i \<le> ilen w"    
   show " x ( inth w i) = i" 
   using 0 1  2 3 
    proof (induct i)
    case 0
    then show ?case by simp
    next
    case (Suc i)
    then show ?case by simp
    qed
   qed
qed 


lemma (in Test1) test_E_F5_2:
 "(
           x ( inth w (0::nat)) = (0::nat) \<and>
           (\<forall>i\<le>ilen w. x ( inth w (i)) = i) \<and>
           x ( inth w (ilen w)) = n) \<longrightarrow> (
           x ( inth w (0::nat)) = (0::nat) \<and>
           (\<forall>i<ilen w. x ( inth w (Suc i)) = Suc (x ( inth w i))) \<and>
           x ( inth w (ilen w)) = n) 
   "
by simp

lemma (in Test1) test_E_F5_3:
 "(
           x ( inth w (0::nat)) = (0::nat) \<and>
           (\<forall>i<ilen w. x ( inth w (Suc i)) = Suc (x ( inth w i))) \<and>
           x ( inth w (ilen w)) = n) =
   (
           x ( inth w (0::nat)) = (0::nat) \<and>
           (\<forall>i\<le>ilen w. x ( inth w (i)) = i) \<and>
           x ( inth w (ilen w)) = n)"
using test_E_F5_1 test_E_F5_2 by auto

lemma (in Test1) test_E_F5_4:
 "(\<exists>x::state \<Rightarrow> nat.
           x ( inth w (0::nat)) = (0::nat) \<and>
           (\<forall>i<ilen w. x ( inth w (Suc i)) = Suc (x ( inth w i))) \<and>
           x ( inth w (ilen w)) = n) =
   (\<exists>x::state \<Rightarrow> nat.
           x ( inth w (0::nat)) = (0::nat) \<and>
           (\<forall>i\<le>ilen w. x ( inth w (i)) = i) \<and>
           x ( inth w (ilen w)) = n)"
by (simp add: Test1.test_E_F5_3)


lemma (in Test1) test_E_F5:
 "\<turnstile> (\<exists>\<exists> v. (F5 v n)) \<longrightarrow> (len n)"
by (auto simp add: Valid_def F5_def exist_state_d_def itl_defs sub_def )
   (metis Test1.test_E_F5_1 nat_le_linear)

subsection \<open>Example 4\<close>

locale Testrev =
 fixes x :: "state \<Rightarrow> nat"
 fixes F1 :: "nat statefun \<Rightarrow> temporal "
 defines "F1 \<equiv> (\<lambda> v. TEMP $v=#0 \<and> skip \<and> v:= $v+#1 )"

lemma (in Testrev) testrev1:
 " (\<sigma> \<Turnstile> F1 (x)) = (ilen \<sigma> = 1 \<and> (x (inth \<sigma> 0)) = 0 \<and> (x (inth \<sigma> 1)) = 1)"
by (simp add: F1_def itl_defs next_assign_d_def, auto)

lemma (in Testrev) testrev2:
 " (\<sigma> \<Turnstile> (F1 (x))\<^sup>r) = (ilen \<sigma> = 1 \<and> (x (inth \<sigma> 0)) = 1 \<and> (x (inth \<sigma> 1)) = 0)"
proof -
 have  "(\<sigma> \<Turnstile> (F1 (x))\<^sup>r) = (\<sigma> \<Turnstile> ($x=#0 \<and> skip \<and> x := $x+#1)\<^sup>r) " 
      by (simp add: F1_def)
 also have  "... = 
          (\<sigma> \<Turnstile> (($x=#0)\<^sup>r \<and> skip\<^sup>r \<and> (x := $x+#1)\<^sup>r))" 
      by (simp add: all_rev_eq)
 also have "... =
          (\<sigma> \<Turnstile> ((!x=#0) \<and> skip \<and> (x! = !x+#1))) "
      using  RRAnd all_rev_eq
      by (simp add: all_rev_eq(1) all_rev_eq(12) all_rev_eq(3) all_rev_eq(8) all_rev_eq(9) 
        next_assign_d_def)
 also have  "... =
          (\<sigma> \<Turnstile> ((x$=#0) \<and> skip \<and> ($x = x$+#1))) "
      by (simp add: itl_defs, auto)
 also have "... =
          (ilen \<sigma> = 1 \<and> (x (inth \<sigma> 0)) = 1 \<and> (x (inth \<sigma> 1)) = 0)"
      by (simp add: itl_defs, auto)
 finally show "(\<sigma> \<Turnstile> (F1 (x))\<^sup>r) = (ilen \<sigma> = 1 \<and> (x (inth \<sigma> 0)) = 1 \<and> (x (inth \<sigma> 1)) = 0)" .
qed


subsection \<open>Example 5\<close>

lemma revnextcurrentfinpenult:
 "\<turnstile>  (v$ = $v)\<^sup>r = (v!=!v)"
proof -
 have 1: "\<turnstile> (v$ = $v)\<^sup>r = ( (v$)\<^sup>r = ($v)\<^sup>r)" by (simp add: rev_fun2)
 have 2: "\<turnstile> ((v$)\<^sup>r =  (v!))" by (simp add: rev_next)
 have 3: "\<turnstile> (($v)\<^sup>r =  (!v))" by (simp add: rev_current) 
 have 4: " \<turnstile>  (( (v$)\<^sup>r = ($v)\<^sup>r) = ( (v!) = (!v) ))" by (metis "1" "2" "3" inteq_reflection)
 from 1 4 show ?thesis by fastforce
qed



end
